package ij.plugin.frame;

import ij.IJ;
import ij.IJEventListener;
import ij.Prefs;
import ij.ImagePlus;
import ij.WindowManager;
import ij.gui.GUI;
import ij.gui.ColorChooser;
import ij.gui.Roi;
import ij.gui.Toolbar;
import ij.process.ColorProcessor;
import ij.plugin.frame.*;

import java.awt.AWTEvent;
import java.awt.BorderLayout;
import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Panel;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.border.BevelBorder;
import javax.swing.border.Border;

/** Implements the Image/Color/Color Picker command. */
public class ColorPicker extends PlugInFrame {
    static final String LOC_KEY = "cp.loc";
    public static ColorPicker instance;
    public static boolean background;
    public static ColorGenerator cg;
    private BevelBorder bb ;//new BevelBorder for LOWER or RAISED type
    private Border grayLBorder = BorderFactory.createBevelBorder(bb.LOWERED, Color.gray, Color.lightGray);
    private Border grayRBorder = BorderFactory.createBevelBorder(bb.RAISED, Color.gray, Color.lightGray);
    public static ColorCanvas colorCanvas;
    private Rectangle fg ;
    private Rectangle bg ;
    private Rectangle spectrumRect;
    private ImagePlus imp;
    private Roi roi;

    public void run(String arg){
        new ColorPicker();
    };

    public ColorPicker() {
            super("CP");
            imp= WindowManager.getCurrentImage();
            if(imp != null)roi=imp.getRoi();
            if (instance!=null) {
                    WindowManager.toFront(instance);
                    return;
            }
            instance = this;
            WindowManager.addWindow(this);
    int colorWidth = 22;
    int colorHeight = 16;
    int columns = 5;
    int rows = 20;
    int width = columns*colorWidth;
    int height = rows*colorHeight;
    addKeyListener(IJ.getInstance());
            setLayout(new BorderLayout());
    cg = new ColorGenerator(width, height, new int[width*height]);
    cg.drawColors(colorWidth, colorHeight, columns, rows);
    if(background)
    cg.refreshBackground();
    colorCanvas = new ColorCanvas(width, height, this, cg);
    Panel panel = new Panel();
    panel.add(colorCanvas);
    add(panel);
            setResizable(false);
            pack();
            Point loc = Prefs.getLocation(LOC_KEY);
            if (loc!=null)
                    setLocation(loc);
            else
                    GUI.center(this);
            setVisible(true);
}
public static void refresh(){
    cg.refreshBackground();
    cg.refreshForeground();
    colorCanvas.repaint();
}

public void close() {
            super.close();
            instance = null;
            Prefs.saveLocation(LOC_KEY, getLocation());
            IJ.notifyEventListeners(IJEventListener.COLOR_PICKER_CLOSED);
    }



class ColorGenerator extends ColorProcessor{
int w, h, cpW, cpH, ccW,ccH, cpControllH, deltaXY;
int[] colors = {0xff0000, 0x00ff00, 0x0000ff, 0xffffff, 0x00ffff, 0xff00ff, 0xffff00, 0x000000};
Roi roi;

public ColorGenerator(int width, int height, int[] pixels) {
    super(width, height, pixels);
}
    void drawColors(int colorWidth, int colorHeight, int columns, int rows) {
    w = colorWidth;
    h = colorHeight;
    cpW=columns*colorWidth;
    cpH=rows*colorHeight;
    setColor(0xffffff);
    setRoi(0, 0, cpW, cpH);
    fill();

    drawRamp();
    drawSpectrum(w, h);

    resetBW();
    cpControllH=4*colorHeight;
    ccH=cpH-cpControllH;
    spectrumRect=new Rectangle(0, 0, cpW, ccH);
    deltaXY=10;
    drawLine(0, ccH, cpW, ccH);
    fg =new Rectangle(deltaXY+2, ccH+2+deltaXY+2, w*2-4, h*2-4);
    bg =new Rectangle((w*2)-deltaXY+2, ccH+2+deltaXY*2+2, w*2-4, h*2-4);

    flipper();

    refreshBackground();
    refreshForeground();
    resetRoi();
}

    void drawColor(int x, int y, Color c) {
        setRoi(x*w, y*h, w, h);
        setColor(c);
        fill();
    }
    void drawColor(int x, int y, int w, int h, Color c) {
        roi = new Roi(x, y, w, h);
        setRoi(roi);
        setColor(c);
        fill();
    }
    void drawBorder(int x, int y, int w, int h) {
        setColor(0x444444);
        drawRect(x-(deltaXY+2), y+deltaXY, w+4, h+4);
        setColor(0x999999);
        drawRect(x-(deltaXY+1), y+deltaXY+1, w+2, h+2);
    }

    public void refreshBackground() {
        //Boundary for Background Selection
        drawBorder((w*2), ccH+deltaXY,(w*2), (h*2));
        drawColor((w*2)-deltaXY, ccH+2+deltaXY*2, w*2, h*2, Toolbar.getBackgroundColor());
        flipper();
    }

    public void refreshForeground() {
        //Boundary for Foreground Selection
        drawBorder(2*deltaXY, ccH,(w*2), (h*2));
        drawColor(deltaXY, ccH+2+deltaXY, w*2, h*2, Toolbar.getForegroundColor());
        flipper();
    }

    void drawSpectrum(double colorWidth, double colorHeight) {
        int x = 1;
        int y = 0;

        Color c;
        float hue, saturation=1f, brightness=1f;
        double w=colorWidth, h=colorHeight;

        for ( x=2; x<10; x++) {
            for ( y=0; y<32; y++) {
                hue = (float)(y/(2*h)-.15);
                if (x<6) {
                    saturation = 1f;
                    brightness = (float)(x*4/w);
                } else {
                    saturation = 1f - ((float)((5-x)*-4/w));
                    brightness = 1f;
                }
                if(x == 5 || x == 6){
                saturation = 1f;
                brightness = 1f;
                        if(y == 3){
                setRoi(x*(int)(w/2), y*(int)(h/2), (int)w, (int)h); //Solid red
                c = Color.getHSBColor(0, saturation, brightness);
                setColor(c);
                fill();continue;
                        }
                        if(y == 10){
                setRoi(x*(int)(w/2), y*(int)(h/2), (int)w, (int)h/2); //Solid yellow
                c = Color.getHSBColor(60f/360f, saturation, brightness);
                setColor(c);
                fill();continue;
                        }
                        if(y == 15){
                setRoi(x*(int)(w/2), y*(int)(h/2), (int)w, (int)h); //Solid green
                c = Color.getHSBColor(120f/360f, saturation, brightness);
                setColor(c);
                fill();continue;
                        }
                        if(y == 21){
                setRoi(x*(int)(w/2), y*(int)(h/2), (int)w, (int)h/2); //Solid cyan
                c = Color.getHSBColor(180f/360f, saturation, brightness);
                setColor(c);
                fill();continue;
                        }
                        if(y == 26){
                setRoi(x*(int)(w/2), y*(int)(h/2), (int)w, (int)h); //Solid blue
                c = Color.getHSBColor(240f/360f, saturation, brightness);
                setColor(c);
                fill();continue;
                        }
                        if(y == 31){
                setRoi(x*(int)(w/2), y*(int)(h/2), 22, (int)h/2); //Solid magenta
                c = Color.getHSBColor(300f/360f, saturation, brightness);
                setColor(c);
                fill();continue;
                        }
                }
                c = Color.getHSBColor(hue, saturation, brightness);
                //IJ.log("x="+x*(int)(w/2));
                setRoi(x*(int)(w/2), y*(int)(h/2), (int)w/2, (int)h/2);
                setColor(c);
                fill();
            }
        }
    }


    void drawSpectrum(double h) {
            Color c;
            for ( int x=5; x<7; x++) {
                    for ( int y=0; y<32; y++) {
                            float hue = (float)(y/(2*h)-.15);
                            c = Color.getHSBColor(hue, 1f, 1f);
                            setRoi(x*(int)(w/2), y*(int)(h/2), (int)w/2, (int)h/2);
                            setColor(c);
                            fill();
                    }
            }
            setRoi(55, 32, 22, 16); //Solid red
            setColor(0xff0000);
            fill();
            setRoi(55, 120, 22, 16); //Solid green
            setColor(0x00ff00);
            fill();
            setRoi(55, 208, 22, 16); //Solid blue
            setColor(0x0000ff);
            fill();
            setRoi(55, 80, 22, 8); //Solid yellow
            setColor(0xffff00);
            fill();
            setRoi(55, 168, 22, 8); //Solid cyan
            setColor(0x00ffff);
            fill();
            setRoi(55, 248, 22, 8); //Solid magenta
            setColor(0xff00ff);
            fill();
    }

void drawRamp() {
    int r,g,b;
    for (int x=0; x<w; x++) {
         for (double y=0; y<(h*16); y++) {
            r = g = b = (byte)y;
            pixels[(int)y*width+x] = 0xff000000 | ((r<<16)&0xff0000) | ((g<<8)&0xff00) | (b&0xff);
        }
    }
}

void resetBW() {   //Paints the Color Reset Button
    setColor(0x000000);
    drawRect(92, 300, 9, 7);
    setColor(0x000000);
    setRoi(88, 297, 9, 7);
    fill();
}

public void flipper() {   //Paints the Flipper Button
    int xa = 90;
    int ya = 272;
    int wa = 18;
    int ha = 18;
    Rectangle flipperRect = new Rectangle(xa-4, ya-4, wa, ha);
    setColor(0x000000);
    drawLine(xa, ya, xa+9, ya+9);//Main Body
    drawLine(xa+1, ya, xa+9, ya+8);
    drawLine(xa, ya+1, xa+8, ya+9);
    Color bgColor=Toolbar.getBackgroundColor();
    setColor(bgColor);
    drawLine(xa, ya, xa, ya+5);//Upper Arrow
    drawLine(xa+1, ya+1, xa+1, ya+6);
    drawLine(xa+2, ya+2, xa+2, ya+3);
    drawLine(xa, ya, xa+5, ya);
    drawLine(xa+1, ya+1, xa+6, ya+1);
    drawLine(xa+2, ya+2, xa+3, ya+2);
    Color fgColor=Toolbar.getForegroundColor();
    setColor(fgColor);
    drawLine(xa+9, ya+9, xa+9, ya+4);//Lower Arrow
    drawLine(xa+8, ya+8, xa+8, ya+3);
    drawLine(xa+7, ya+7, xa+7, ya+6);
    drawLine(xa+9, ya+9, xa+4, ya+9);
    drawLine(xa+8, ya+8, xa+3, ya+8);
    drawLine(xa+7, ya+7, xa+6, ya+7);
    drawRect(xa-4, ya-4, wa, ha);
    drawRoi(new Roi(flipperRect));
}
public void flipper(Color roiColor) {   //Paints the Flipper Button
    int xa = 90;
    int ya = 272;
    int wa = 18;
    int ha = 18;
    Rectangle flipperRect = new Rectangle(xa-4, ya-4, wa, ha);
    setColor(roiColor);
    drawRect(xa-4, ya-4, wa, ha);
    //drawRoi(new Roi(flipperRect));
}

}

public class ColorCanvas extends Canvas implements MouseListener, MouseMotionListener{
    int width, height;
    Vector colors;
    long mouseDownTime;
    ColorGenerator ip;
    Frame frame;
    Image image;
        boolean fobg;
        Rectangle finalRect;

    public ColorCanvas(int width, int height, Frame frame, ColorGenerator ip) {
            this.width=width; this.height=height;
            this.frame = frame;
            this.ip = ip;
            image=ip.createImage();
            addMouseListener(this);
            addMouseMotionListener(this);
            addKeyListener(IJ.getInstance());
            setSize(width, height);
    }

    public Dimension getPreferredSize() {
            return new Dimension(width, height);
    }

    public void update(Graphics g) {
            paint(g);
    }

    public void paint(Graphics g) {
            g.drawImage(image, 0, 0, this);
    }

    public void mousePressed(MouseEvent e) {
            //IJ.log("mousePressed "+e);
            int x = e.getX();
            int y = e.getY();
            ip.setLineWidth(1);
            if (Toolbar.getToolId()==Toolbar.DROPPER)
            IJ.setTool(Toolbar.RECTANGLE );

            Rectangle colorRect = new Rectangle(0, 0, 110, 256);
            Rectangle controllRect = new Rectangle(0, 256, 110, 320);
            Rectangle flipperRect = new Rectangle(86, 268, 18, 18);
            Rectangle resetRect = new Rectangle(86, 294, 18, 18);
            Rectangle foreground1Rect = new Rectangle(9, 266, 45, 10);
            Rectangle foreground2Rect = new Rectangle(9, 276, 23, 25);
            Rectangle overlappingRect = new Rectangle(33, 276, 23, 25);
            Rectangle background1Rect = new Rectangle(33, 302, 45, 10);
            Rectangle background2Rect = new Rectangle(56, 277, 23, 25);

            long difference = System.currentTimeMillis()-mouseDownTime;
            boolean doubleClick = (difference<=250);
            mouseDownTime = System.currentTimeMillis();

            if (flipperRect.contains(x, y)) {
                    Color c = Toolbar.getBackgroundColor();
                    Toolbar.setBackgroundColor(Toolbar.getForegroundColor());
                    Toolbar.setForegroundColor(c);
                        setControllRectColor(background);
                    return;
                 } else if(resetRect.contains(x,y)) {
                    Toolbar.setForegroundColor(new Color(0x000000));
                    Toolbar.setBackgroundColor(new Color(0xffffff));
                        setControllRectColor(background);
                    return;
            } else if ((background1Rect.contains(x,y)) || (background2Rect.contains(x,y))) {
                    background = true;
                    setDrawingColor(x, y, background);
            } else if ((foreground1Rect.contains(x,y)) || (foreground2Rect.contains(x,y))) {
                    background = false;
                    setDrawingColor(x, y, background);
            } else {
                    if (overlappingRect.contains(x,y)){
                        setDrawingColor(x, y, fobg);
                    }
                    else if ((colorRect.contains(x,y)) || (flipperRect.contains(x, y) || resetRect.contains(x,y))){
                        setDrawingColor(x, y, background);
                        setControllRectColor(background);
                        return;
                    }
                    else if (!(flipperRect.contains(x, y) || resetRect.contains(x,y)) && controllRect.contains(x,y)){
                        return;
                    }
            }
            finalRect = background ? bg : fg;
            if (dialogItemChanged(finalRect, e)) {
                    ip.refreshForeground();
                    ip.refreshBackground();
            } else {
                    ip.refreshBackground();
                    ip.refreshForeground();
            }
            repaint();
    }

    public void mouseMoved(MouseEvent e) {
            int x = e.getX();
            int y = e.getY();
            Point cp = e.getPoint();
            if(spectrumRect.contains(cp)
                        || fg.contains(cp)
                          || bg.contains(cp)){
            int p = ip.getPixel(x, y);
            int r = (p&0xff0000)>>16;
            int g = (p&0xff00)>>8;
            int b = p&0xff;
            Color mouseMoveColor=new Color(r, g, b);
            ip.flipper(mouseMoveColor);repaint();
            if(roi != null){roi.setColor(mouseMoveColor);
            if (imp!=null && imp.getRoi()!=null) imp.draw();
            }
            IJ.showStatus("red="+pad(r)+", green="+pad(g)+", blue="+pad(b));
            }

    }

    String pad(int n) {
            String str = ""+n;
            while (str.length()<3)
            str = "0" + str;
            return str;
    }

    void setControllRectColor(boolean background){
        if (background) {
            ip.refreshForeground();
            ip.refreshBackground();
         } else {
            ip.refreshBackground();
            ip.refreshForeground();
        }
        repaint();
    }

    void setDrawingColor(int x, int y, boolean setBackground) {
            int p = ip.getPixel(x, y);
            int r = (p&0xff0000)>>16;
            int g = (p&0xff00)>>8;
            int b = p&0xff;
            Color c = new Color(r, g, b);
            if (setBackground) {
                    Toolbar.setBackgroundColor(c);
                    if (Recorder.record)
                            Recorder.record("setBackgroundColor", c.getRed(), c.getGreen(), c.getBlue());
            } else {
                    Toolbar.setForegroundColor(c);
                    if (Recorder.record)
                            Recorder.record("setForegroundColor", c.getRed(), c.getGreen(), c.getBlue());
            }
    }

    void editColor(AWTEvent awte) {
        MouseEvent e = (MouseEvent)awte;
        int x = e.getX();
        int y = e.getY();
        Color c  = background?Toolbar.getBackgroundColor():Toolbar.getForegroundColor();
        if(awte!=null){
            ColorChooser cc = new ColorChooser((background?"Background":"Foreground")+" Color", c, false);
            c = cc.getColor();
        }
            if (background)
                    Toolbar.setBackgroundColor(c);
            else
                    Toolbar.setForegroundColor(c);
    }

    public void refreshColors() {
            ip.refreshBackground();
            ip.refreshForeground();
            repaint();
    }

    public void mouseReleased(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}
    public void mouseClicked(MouseEvent e) {}
    public void mouseEntered(MouseEvent e) {}
    public void mouseDragged(MouseEvent e) {}

        public boolean dialogItemChanged(Shape roi, AWTEvent e) {
                if(roi == fg){
                        editColor(e);
                        return fobg=false;
                        }
                if(roi == bg){
                        editColor(e);
                        return fobg=true;
                        }
                return fobg;
        }

}


}
